// ---------------------------------------------------------------------------
// - Unicode.cpp                                                             -
// - standard object library - unicode functions class implementation        -
// ---------------------------------------------------------------------------
// - This program is free software;  you can redistribute it  and/or  modify -
// - it provided that this copyright notice is kept intact.                  -
// -                                                                         -
// - This program  is  distributed in  the hope  that it will be useful, but -
// - without  any  warranty;  without  even   the   implied    warranty   of -
// - merchantability or fitness for a particular purpose.  In no event shall -
// - the copyright holder be liable for any  direct, indirect, incidental or -
// - special damages arising in any way out of the use of this software.     -
// ---------------------------------------------------------------------------
// - copyright (c) 1999-2006 amaury darsch                                   -
// ---------------------------------------------------------------------------

#include "Ascii.hpp"
#include "Unicode.hpp"
#include "Exception.hpp"
#include "ccnv.hpp"

namespace afnix {

  // convert a unicode character to a native character if possible

  char Unicode::tochar (const t_quad value) {
    // check for 8 bit range
    if ((value & 0xFFFFFF00) != 0x00000000)
      throw Exception ("unicode-error", "cannot convert unicode character");
    // map the character
    char result = (char) (value & 0x000000FF);
    return result;
  }

  // convert a native character to a unicode character

  t_quad Unicode::toquad (const char value) {
    t_quad result = value;
    return result & 0x000000ff;
  }

  // convert a string representation to a character

  t_quad Unicode::toquad (const String& value) {
    long slen = value.length ();
    // check for single character
    if (slen == 1) {
      t_quad result = (t_quad) value[0];
      return result;
    }
    // check for ascii representation
    if ((slen > 2) && (value[0] == '\'')) {
      t_quad result = (t_quad) Ascii::tochar (value);
      return result;
    }
    // check for unicode representation
    if ((slen > 2) && (value[0] == 'U') && (value[1] == '+')) {
      // format the string
      String format = "0x";
      format += value.rsubstr (2);
      // convert to quad
      bool  status = false;
      const char* data = format.tochar ();
      t_quad result = (t_quad) c_atoll (data, status);
      delete [] data;
      if (status == false) {
	throw Exception ("literal-error",
			 "illegal unicode string representation", value);
      }
      return result;
    }
    // invalid format
    throw Exception ("format-error",
		     "illegal unicode string representation", value);
  }

  // convert a unicode character value to a string

  String Unicode::tostring (const t_quad value) {
    // check for an ascii character
    if ((value & 0xFFFFFF00) == 0x00000000) {
      char cval = (char) (value & 0x000000FF);
      String result = Ascii::tostring (cval);
      return result;
    }
    // we are outside the ascii range, so use the unicode representation
    char* data = c_ltoh (value, false);
    if (data == nilp) {
      throw Exception ("unicode-error", "cannot convert unicode character");
    }
    String result = "U+";
    result += data;
    delete [] data;
    return result;
  }

  // convert a native character value to a literal string

  String Unicode::toliteral (const t_quad value) {
    String result;
    if (Unicode::isascii (value) == true) {
      char cval = (char) (value & 0x000000FF);
      result += '\'';
      result += cval;
      result += '\'';
    } else {
      result += '"';
      result += Unicode::tostring (value);
      result += '"';
    }
    return result;
  }

  // get the size of unicode array

  long Unicode::strlen (const t_quad* s) {
    // check for nil string
    if (s == nilp) return 0;
    // compute length
    long result = 0;
    while (*s++ != nilq) result++;
    return result;
  }

  // compare two strings and returns true if they are equals.

  bool Unicode::strcmp (const t_quad* s1, const char* s2) {
    // nil case first
    if ((s1 == nilp) &&  (s2 == nilp)) return true;
    if ((s1 == nilp) &&  (s2 != nilp) && (*s2 == nilc)) return true;
    if ((s1 == nilp) &&  (s2 != nilp) && (*s2 != nilc)) return false;
    if ((s1 != nilp) && (*s1 == nilq) &&  (s2 == nilp)) return true;
    if ((s1 != nilp) && (*s1 != nilq) &&  (s2 == nilp)) return false;
    // check first character for fast compare
    if (*s1 != (t_quad) *s2) return false;
    // normal compare now
    while (*s1 != nilq) {
      if (*s2 == nilc) break;
      if (*s1++ != (t_quad) *s2++) return false;
    }
    return (*s1 == (t_quad) *s2) ? true : false;
  }

  // compare two strings and returns true if they are equals.

  bool Unicode::strcmp (const t_quad* s1, const t_quad* s2) {
    // nil case first
    if ((s1 == nilp) &&  (s2 == nilp)) return true;
    if ((s1 == nilp) &&  (s2 != nilp) && (*s2 == nilq)) return true;
    if ((s1 == nilp) &&  (s2 != nilp) && (*s2 != nilq)) return false;
    if ((s1 != nilp) && (*s1 == nilq) &&  (s2 == nilp)) return true;
    if ((s1 != nilp) && (*s1 != nilq) &&  (s2 == nilp)) return false;
    // check first character for fast compare
    if (*s1 != *s2) return false;
    // normal compare now
    while (*s1 != nilq) {
      if (*s2 == nilq) break;
      if (*s1++ != *s2++) return false;
    }
    return (*s1 == *s2) ? true : false;
  }

  // compare two strings upto n characters

  bool Unicode::strncmp (const t_quad* s1, const char* s2, const long size) {
    // nil case compare
    if (size == 0) return true;
    long len1 = Unicode::strlen (s1);
    long len2 = Ascii::strlen   (s2);
    if ((len1 == 0) || (len2 == 0)) return false;
    // normal compare
    for (long i = 0; i < size; i++) {
      if ((s1[i] == nilq) && (s2[i] == nilc)) return false;
      if (s1[i] != (t_quad) s2[i]) return false;
    }
    return true;
  }

  // compare two strings upto n characters

  bool Unicode::strncmp (const t_quad* s1, const t_quad* s2, const long size) {
    // nil case compare
    if (size == 0) return true;
    long len1 = Unicode::strlen (s1);
    long len2 = Unicode::strlen (s2);
    if ((len1 == 0) || (len2 == 0)) return false;
    // normal compare
    for (long i = 0; i < size; i++) {
      if ((s1[i] == nilq) && (s2[i] == nilq)) return false;
      if (s1[i] != s2[i]) return false;
    }
    return true;
  }

  // compare two strings - less than operator

  bool Unicode::strlth (const t_quad* s1, const char* s2) {
    // nil case first
    if ((s1 == nilp) &&  (s2 == nilp)) return false;
    if ((s1 == nilp) &&  (s2 != nilp) && (*s2 == nilc)) return false;
    if ((s1 == nilp) &&  (s2 != nilp) && (*s2 != nilc)) return true;
    if ((s1 != nilp) &&  (s2 == nilp) && (*s1 == nilq)) return false;
    if ((s1 != nilp) &&  (s2 == nilp) && (*s1 != nilq)) return false;
    // compare without equal
    while (*s1 != nilq) {
      if (*s1   < (t_quad) *s2)   return true;
      if (*s1++ > (t_quad) *s2++) return false;
    }
    return false;
  }


  // compare two strings - less than operator

  bool Unicode::strlth (const t_quad* s1, const t_quad* s2) {
    // nil case first
    if ((s1 == nilp) &&  (s2 == nilp)) return false;
    if ((s1 == nilp) &&  (s2 != nilp) && (*s2 == nilq)) return false;
    if ((s1 == nilp) &&  (s2 != nilp) && (*s2 != nilq)) return true;
    if ((s1 != nilp) &&  (s2 == nilp) && (*s1 == nilq)) return false;
    if ((s1 != nilp) &&  (s2 == nilp) && (*s1 != nilq)) return false;
    // compare without equal
    while (*s1 != nilq) {
      if (*s1   < *s2)   return true;
      if (*s1++ > *s2++) return false;
    }
    return false;
  }

  // compare two strings - less equal operator

  bool Unicode::strleq (const t_quad* s1, const char* s2) {
    // nil case first
    if ((s1 == nilp) &&  (s2 == nilp)) return true;
    if ((s1 == nilp) &&  (s2 != nilp) && (*s2 == nilc)) return true;
    if ((s1 == nilp) &&  (s2 != nilp) && (*s2 != nilc)) return true;
    if ((s1 != nilp) &&  (s2 == nilp) && (*s1 == nilq)) return true;
    if ((s1 != nilp) &&  (s2 == nilp) && (*s1 != nilq)) return false;
    // compare with equal
    while (*s1 != nilq) {
      if (*s1   < (t_quad) *s2)   return true;
      if (*s1++ > (t_quad) *s2++) return false;
    }
    return true;
  }

  // compare two strings - less equal operator

  bool Unicode::strleq (const t_quad* s1, const t_quad* s2) {
    // nil case first
    if ((s1 == nilp) &&  (s2 == nilp)) return true;
    if ((s1 == nilp) &&  (s2 != nilp) && (*s2 == nilq)) return true;
    if ((s1 == nilp) &&  (s2 != nilp) && (*s2 != nilq)) return true;
    if ((s1 != nilp) &&  (s2 == nilp) && (*s1 == nilq)) return true;
    if ((s1 != nilp) &&  (s2 == nilp) && (*s1 != nilq)) return false;
    if ((s1 != nilp) &&  (s2 == nilp) && (*s1 != nilq)) return false;
    // compare with equal
    while (*s1 != nilq) {
      if (*s1   < *s2)   return true;
      if (*s1++ > *s2++) return false;
    }
    return true;
  }

  // convert an ascii character to an unicode array

  t_quad* Unicode::strmak (const char value) {
    t_quad* result = new t_quad[2];
    result[0] = (t_quad) value;
    result[1] = nilq;
    return result;
  }

  // convert a unicode character to an unicode array

  t_quad* Unicode::strmak (const t_quad value) {
    t_quad* result = new t_quad[2];
    result[0] = value;
    result[1] = nilq;
    return result;
  }

  // convert an ascii string to an unicode array

  t_quad* Unicode::strdup (const char* s) {
    // check for null string if the length is null
    long len = Ascii::strlen (s);
    // allocate a new string with new so delete can be used
    t_quad* result = new t_quad[len + 1];
    if (len > 0) {
      for (long i = 0; i < len + 1; i++) result[i] = (t_quad) s[i];
    }
    result[len] = nilq;
    return result;
  }

  // convert a unicode string to an unicode array

  t_quad* Unicode::strdup (const t_quad* s) {
    // check for null string if the length is null
    long len = Unicode::strlen (s);
    // allocate a new string with new so delete can be used
    t_quad* result = new t_quad[len + 1];
    if (len > 0) {
      for (long i = 0; i < len + 1; i++) result[i] = s[i];
    }
    result[len] = nilq;
    return result;
  }

  // copy a c-string from a source to a unicode destination

  void Unicode::strcpy (t_quad* dst, const char* src) {
    // standard check as usual
    if (dst == nilp) return;
    // get the length and copy
    long len = Ascii::strlen (src);
    if (len > 0) {
      for (long i = 0; i < len + 1; i++) dst[i] = (t_quad) src[i];
    }
    dst[len] = nilq;
  }

  // copy a unicode string from a source to a unicode destination

  void Unicode::strcpy (t_quad* dst, const t_quad* src) {
    // standard check as usual
    if (dst == nilp) return;
    // get the length and copy
    long len = Unicode::strlen (src);
    if (len > 0) {
      for (long i = 0; i < len + 1; i++) dst[i] = src[i];
    }
    dst[len] = nilq;
  }

  // concatenate a c-string with a unicode string

  void Unicode::strcat (t_quad* dst, const char* src) {
    // standard check as usual
    if (dst == nilp) return;
    // get length and position
    long len = Ascii::strlen   (src);
    long pos = Unicode::strlen (dst);
    if (len > 0) {
      for (long i = 0; i < len + 1; i++) dst[pos + i] = (t_quad) src[i];
    }
    dst[pos+len] = nilq;
  }

  // concatenate a unicode string with a unicode string

  void Unicode::strcat (t_quad* dst, const t_quad* src) {
    // standard check as usual
    if (dst == nilp) return;
    // get length and position
    long len = Unicode::strlen (src);
    long pos = Unicode::strlen (dst);
    if (len > 0) {
      for (long i = 0; i < len + 1; i++) dst[pos + i] = src[i];
    }
    dst[pos+len] = nilq;
  }

  // remove the leading blank and tab and return a new string

  t_quad* Unicode::rmlead (const char* s) {
    // get the length and remove
    long len = Ascii::strlen (s);
    if (len != 0) {
      // remove leading blank
      while ((*s != nilc) && ((*s == blkc) || (*s == tabc))) s++;
    }
    // now copy
    return Unicode::strdup (s);
  }
 
  // remove the leading blank and tab and return a new string

  t_quad* Unicode::rmlead (const t_quad* s) {
    // get the length and remove
    long len = Unicode::strlen (s);
    if (len != 0) {
      // remove leading blank
      while ((*s != nilq) && ((*s == blkq) || (*s == tabq))) s++;
    }
    // now copy
    return Unicode::strdup (s);
  }

  // remove the trailing blank from a string and return a new string

  t_quad* Unicode::rmtrail (const char* s) {
    // get the length and check
    long len = Ascii::strlen (s);
    if (len == 0) return Unicode::strdup (s);
    char* sbuf = Ascii::strdup (s);
    char* end  = sbuf + len - 1;
    // remove trailing blank
    while ((end != s) && ((*end == blkc) || (*end == tabc))) *end-- = nilc;
    // now copy and return
    t_quad* result = Unicode::strdup (sbuf);
    delete [] sbuf;
    return result;
  }

  // remove the trailing blank from a string and return a new string

  t_quad* Unicode::rmtrail (const t_quad* s) {
    // get the length and check
    long len = Unicode::strlen (s);
    if (len == 0) return Unicode::strdup (s);
    t_quad* sbuf = Unicode::strdup (s);
    t_quad* end  = sbuf + len - 1;
    // remove trailing blank
    while ((end != s) && ((*end == blkq) || (*end == tabq))) *end-- = nilq;
    // now copy and return
    t_quad* result = Unicode::strdup (sbuf);
    delete [] sbuf;
    return result;
  }

  // the lower case extension factor
  static long TOLOWER_MAX = 1;

  // convert a character from upper case to lower case
  static long unicode_tolower (t_quad* dst, const t_quad src) {
    // plane 0
    // character A-Z to a-z
    if ((src >= 0x00000041) && (src <= 0x0000005A)) {
      dst[0] = (src - 0x00000041) + 0x00000061;
      return 1;
    }
    // micro sign
    if (src == 0x000000B5) {
      dst[0] = 0x000003BC;
      return 1;
    }
    // check multiplication sign
    if (src == 0x000000D7) {
      dst[0] = src;
      return 1;
    }
    // character U+00C0 to U+00DE
    if ((src >= 0x000000C0) && (src <= 0x000000DE)) {
      dst[0] = (src - 0x000000C0) + 0x000000E0;
      return 1;
    }
    // plane 1
    // check for latin capital Y with diaeresis
    if (src == 0x00000178) {
      dst[0] = 0x000000FF;
      return 1;
    }
    // by default do not change anything
    dst[0] = src;
    return 1;
  }

  // convert an ascii string to lower case

  t_quad* Unicode::tolower (const char* s) {
    // check for length
    if (s == nilp) return nilp;
    long len = Ascii::strlen (s);
    // allocate and convert
    long    size = len * TOLOWER_MAX + 1;
    t_quad* sbuf = new t_quad[size];
    long    sidx = 0;
    t_quad  sdst[TOLOWER_MAX];
    for (long i = 0; i < len; i++) {
      long cnvs = unicode_tolower (sdst, (t_quad) s[i]);
      for (long j = 0; j < cnvs; j++) sbuf[sidx++] = sdst[j];
    }
    // mark end of string
    sbuf[sidx] = nilq;
    // copy and clean
    t_quad* result = Unicode::strdup (sbuf);
    delete [] sbuf;
    return result;
  }

  // convert an ascii string to lower case

  t_quad* Unicode::tolower (const t_quad* s) {
    // check for length
    if (s == nilp) return nilp;
    long len = Unicode::strlen (s);
    // allocate and convert
    long    size = len * TOLOWER_MAX + 1;
    t_quad* sbuf = new t_quad[size];
    long    sidx = 0;
    t_quad  sdst[TOLOWER_MAX];
    for (long i = 0; i < len; i++) {
      long cnvs = unicode_tolower (sdst, s[i]);
      for (long j = 0; j < cnvs; j++) sbuf[sidx++] = sdst[j];
    }
    // mark end of string
    sbuf[sidx] = nilq;
    // copy and clean
    t_quad* result = Unicode::strdup (sbuf);
    delete [] sbuf;
    return result;
  }

  // the lower case extension factor
  static long TOUPPER_MAX = 2;

  // convert a character from lowercase to upper case

  static long unicode_toupper (t_quad* dst, const t_quad src) {
    // plane 0
    // character a-z to A-Z
    if ((src >= 0x00000061) && (src <= 0x0000007A)) {
      dst[0] = (src - 0x00000061) + 0x00000041;
      return 1;
    }
    // latin small letter sharp S (german) --> SS
    if (src == 0x000000DF) {
      dst[0] = 0x00000073;
      dst[1] = 0x00000073;
      return 2;
    }
    // check division sign
    if (src == 0x000000F7) {
      dst[0] = src;
      return 1;
    }
    // character U+00E0 to U+00FE
    if ((src >= 0x000000E0) && (src <= 0x000000FE)) {
      dst[0] = (src - 0x000000E0) + 0x000000C0;
      return 1;
    }
    // check small letter Y with diaeresis
    if (src == 0x000000FF) {
      dst[0] = 0x00000178;
      return 1;
    }
    // by default do not change anything
    dst[0] = src;
    return 1;
  }

  // convert an ascii string to upper case

  t_quad* Unicode::toupper (const char* s) {
    // check for length
    if (s == nilp) return nilp;
    long len = Ascii::strlen (s);
    // allocate and convert
    long    size = len * TOUPPER_MAX + 1;
    t_quad* sbuf = new t_quad[size];
    long    sidx = 0;
    t_quad  sdst[TOUPPER_MAX];
    for (long i = 0; i < len; i++) {
      long cnvs = unicode_toupper (sdst, (t_quad) s[i]);
      for (long j = 0; j < cnvs; j++) sbuf[sidx++] = sdst[j];
    }
    // mark end of string
    sbuf[sidx] = nilq;
    // copy and clean
    t_quad* result = Unicode::strdup (sbuf);
    delete [] sbuf;
    return result;
  }

  // convert an unicode string to upper case

  t_quad* Unicode::toupper (const t_quad* s) {
    // check for length
    if (s == nilp) return nilp;
    long len = Unicode::strlen (s);
    // allocate and convert
    long    size = len * TOUPPER_MAX + 1;
    t_quad* sbuf = new t_quad[size];
    long    sidx = 0;
    t_quad  sdst[TOUPPER_MAX];
    for (long i = 0; i < len; i++) {
      long cnvs = unicode_toupper (sdst, s[i]);
      for (long j = 0; j < cnvs; j++) sbuf[sidx++] = sdst[j];
    }
    // mark end of string
    sbuf[sidx] = nilq;
    // copy and clean
    t_quad* result = Unicode::strdup (sbuf);
    delete [] sbuf;
    return result;
  }

  // return true if the character is a lower character

  bool Unicode::islower (const t_quad value) {
    // plane 0
    // character a-z to A-Z
    if ((value >= 0x00000061) && (value <= 0x0000007A)) return true;
    // latin small letter sharp S (german) --> SS
    if (value == 0x000000DF) return true;
    // character U+00E0 to U+00FE
    if ((value >= 0x000000E0) && (value <= 0x000000FE)) return true;
    // check small letter Y with diaeresis
    if (value == 0x000000FF) return true;
    // not a lower character
    return false;
  }

  // return true if the character is an upper character

  bool Unicode::isupper (const t_quad value) {
    // plane 0
    // character A-Z
    if ((value >= 0x00000041) && (value <= 0x0000005A)) return true;
    // character U+00C0 to U+00DE
    if ((value >= 0x000000C0) && (value <= 0x000000DE)) return true;
    // plane 1
    // check for latin capital Y with diaeresis
    if (value == 0x00000178) return true;
    // not an upper character
    return false;
  }

  // return true if the unicode character is a letter

  bool Unicode::isletter (const t_quad value) {
    // plane 0
    // ascii characters
    if ((value >= (t_quad) 'a') && (value <= (t_quad) 'z')) return true;
    if ((value >= (t_quad) 'A') && (value <= (t_quad) 'Z')) return true;
    // check multiplication sign
    if (value == 0x000000D7) return false;
    // character U+00C0 to U+00DE
    if ((value >= 0x000000C0) && (value <= 0x000000DE)) return true;
    // check division sign
    if (value == 0x000000F7) return false;
    // character U+00E0 to U+00FF
    if ((value >= 0x000000E0) && (value <= 0x000000FF)) return true;
    // by default it is not a letter
    return false;
  }

  // return true if the unicode character is a digit

  bool Unicode::isdigit (const t_quad value) {
    if ((value >= (t_quad) '0') && (value <= (t_quad) '9')) return true;
    return false;
  }

  // return true if the unicode character is an alpha-numeric character

  bool Unicode::isalpha (const t_quad value) {
    // check for a digit
    if (Unicode::isdigit (value) == true) return true;
    // check for letter
    if (Unicode::isletter (value) == true) return true;
    // not alpha
    return false;
  }

  // return true if the unicode character is a blank or tab
  
  bool Unicode::isblank (const t_quad value) {
    if ((value == blkq) || (value == tabq)) return true;
    return false;
  }

  // return true if the unicode character is an ascii character
  
  bool Unicode::isascii (const t_quad value) {
    if ((value & 0xFFFFFF80) == 0x00000000) return true;
    return false;
  }

  // return true if the unicode character is a latin character
  
  bool Unicode::islatin (const t_quad value) {
    if ((value & 0xFFFFFF00) == 0x00000000) return true;
    return false;
  }

  // return true if the unicode character is an hexadecimal character
  
  bool Unicode::ishexa (const t_quad value) {
    if ((value >= (t_quad) '0') && (value <= (t_quad) '9')) return true;
    if ((value >= (t_quad) 'a') && (value <= (t_quad) 'f')) return true;
    if ((value >= (t_quad) 'A') && (value <= (t_quad) 'F')) return true;
    return false;
  }

  // return true if the character is an afnix constituent

  bool Unicode::isafnix (const t_quad value) {
    // check for an alhpa character
    if (isalpha (value) == true) return true;
    // check for other constituents
    if (value == (t_quad) '.') return true;
    if (value == (t_quad) '+') return true;
    if (value == (t_quad) '-') return true;
    if (value == (t_quad) '*') return true;
    if (value == (t_quad) '/') return true;
    if (value == (t_quad) '!') return true;
    if (value == (t_quad) '=') return true;
    if (value == (t_quad) '.') return true;
    if (value == (t_quad) '>') return true;
    if (value == (t_quad) '<') return true;
    if (value == (t_quad) '?') return true;
    return false;
  }

  // encode a unicode character in UTF-8

  char* Unicode::encode (const t_quad c) {
    return Unicode::encode (&c, 1);
  }

  // encode a unicode string in UTF-8

  char* Unicode::encode (const t_quad* s) {
    // get the size and encode
    long size = Unicode::strlen (s);
    return encode (s, size);
  }

  // encode a unicode string in UTF-8
  
  char* Unicode::encode (const t_quad* s, const long size) {
    // check the size
    if (size <= 0) return nilp;
    // allocate the character buffer
    char* buf = new char[size*6+1];
    long  idx = 0;
    // loop in the buffer and encode
    for (long i = 0; i < size; i++) {
      // get the character value
      t_quad value = s[i];
      // encode the value
      if (value < 0x00000080) {
	buf[idx++] = (char) value;
      } else if (value < 0x00000800) {
	buf[idx++] = (char) (0x000000C0 | ((value >> 6)  & 0x0000001F));
	buf[idx++] = (char) (0x00000080 |  (value        & 0x0000003F));
      } else if (value < 0x00010000) {
	buf[idx++] = (char) (0x000000E0 | ((value >> 12) & 0x0000000F));
	buf[idx++] = (char) (0x00000080 | ((value >> 6)  & 0x0000003F));
	buf[idx++] = (char) (0x00000080 |  (value        & 0x0000003F));
      } else if (value < 0x00200000) {
	buf[idx++] = (char) (0x000000F0 | ((value >> 18) & 0x00000007));
	buf[idx++] = (char) (0x00000080 | ((value >> 12) & 0x0000003F));
	buf[idx++] = (char) (0x00000080 | ((value >> 6)  & 0x0000003F));
	buf[idx++] = (char) (0x00000080 |  (value        & 0x0000003F));
      } else if (value < 0x04000000) {
	buf[idx++] = (char) (0x000000F8 | ((value >> 24) & 0x00000003));
	buf[idx++] = (char) (0x00000080 | ((value >> 18) & 0x0000003F));
	buf[idx++] = (char) (0x00000080 | ((value >> 12) & 0x0000003F));
	buf[idx++] = (char) (0x00000080 | ((value >> 6)  & 0x0000003F));
	buf[idx++] = (char) (0x00000080 |  (value        & 0x0000003F));
      } else if (value < 0x80000000) {
	buf[idx++] = (char) (0x000000FC | ((value >> 30) & 0x00000001));
	buf[idx++] = (char) (0x00000080 | ((value >> 24) & 0x0000003F));
	buf[idx++] = (char) (0x00000080 | ((value >> 18) & 0x0000003F));
	buf[idx++] = (char) (0x00000080 | ((value >> 12) & 0x0000003F));
	buf[idx++] = (char) (0x00000080 | ((value >> 6)  & 0x0000003F));
	buf[idx++] = (char) (0x00000080 |  (value        & 0x0000003F));
      } else {
	throw Exception ("encode-error", 
			 "invalid character to encode in utf-8 mode");
      }
    }
    // add the nil character
    buf[idx++] = nilc;
    // here we are
    return buf;
  }

  // decode a unicode buffer

  t_quad* Unicode::decode (const char* s) {
    // get the size and decode
    long size = Ascii::strlen (s);
    return Unicode::decode (s, size); 
  }

  // decode a unicode buffer by size

  t_quad* Unicode::decode (const char* s, const long size) {
    // check the size
    if (size <= 0) return nilp;
    // allocate the quad buffer
    t_quad* buf = new t_quad[size+1];
    long    idx = 0;
    for (long i = 0; i < size; i++) {
      // read first byte
      t_byte b1 = (t_byte) s[i];
      // 1 byte mode
      if (b1 < 0x80) {
	buf[idx++] = (t_quad) b1;
	continue;
      }
      // 2 bytes mode
      if (b1 < 0xE0) {
	buf[idx] = ((t_quad) (b1 & 0x3F)) << 6;
	if (++i >= size) {
	  throw Exception ("decode-error", 
			   "invalid eos while reading utf-8 sequence");
	}
	t_byte b2 = (t_byte) s[i];
	if ((b2 & 0x80) != 0x80) {
	  throw Exception ("decode-error", 
			   "invalid byte while reading utf-8 sequence");
	}
	buf[idx] |= (t_quad) (b2 & 0x3F);
	if (buf[idx++] < 0x00000080) {
	  throw Exception ("decode-error", "invalid long utf-8 sequence");
	}
	continue;
      }
      // 3 bytes mode
      if (b1 < 0xF0) {
	buf[idx] = ((t_quad) (b1 & 0x0F)) << 12;
	if (++i >= size) {
	  throw Exception ("decode-error", 
			   "invalid eos while reading utf-8 sequence");
	}
	t_byte b2 = (t_byte) s[i];
	if ((b2 & 0x80) != 0x80) {
	  throw Exception ("decode-error", 
			   "invalid byte while reading utf-8 sequence");
	}
	buf[idx] |= ((t_quad) (b2 & 0x3F)) << 6;
	if (++i >= size) {
	  throw Exception ("decode-error", 
			   "invalid eos while reading utf-8 sequence");
	}
	t_byte b3 = (t_byte) s[i];
	if ((b3 & 0x80) != 0x80) {
	  throw Exception ("decode-error", 
			   "invalid byte while reading utf-8 sequence");
	}
	buf[idx] |= (t_quad) (b3 & 0x3F);
	if (buf[idx++] < 0x00000800) {
	  throw Exception ("decode-error", "invalid long utf-8 sequence");
	}
	continue;
      }
      // 4 bytes mode
      if (b1 < 0xF8) {
	buf[idx] = ((t_quad) (b1 & 0x07)) << 18;
	if (++i >= size) {
	  throw Exception ("decode-error", 
			   "invalid eos while reading utf-8 sequence");
	}
	t_byte b2 = (t_byte) s[i];
	if ((b2 & 0x80) != 0x80) {
	  throw Exception ("decode-error", 
			   "invalid byte while reading utf-8 sequence");
	}
	buf[idx] |= ((t_quad) (b2 & 0x3F)) << 12;
	if (++i >= size) {
	  throw Exception ("decode-error", 
			   "invalid eos while reading utf-8 sequence");
	}
	t_byte b3 = (t_byte) s[i];
	if ((b3 & 0x80) != 0x80) {
	  throw Exception ("decode-error", 
			   "invalid byte while reading utf-8 sequence");
	}
	buf[idx] |= (t_quad) (b3 & 0x3F) << 6;
	if (++i >= size) {
	  throw Exception ("decode-error", 
			   "invalid eos while reading utf-8 sequence");
	}
	t_byte b4 = (t_byte) s[i];
	if ((b4 & 0x80) != 0x80) {
	  throw Exception ("decode-error", 
			   "invalid byte while reading utf-8 sequence");
	}
	buf[idx] |= (t_quad) (b4 & 0x3F);
	if (buf[idx++] < 0x00010000) {
	  throw Exception ("decode-error", "invalid long utf-8 sequence");
	}
	continue;
      }
      // 5 bytes mode
      if (b1 < 0xFC) {
	buf[idx] = ((t_quad) (b1 & 0x03)) << 24;
	if (++i >= size) {
	  throw Exception ("decode-error", 
			   "invalid eos while reading utf-8 sequence");
	}
	t_byte b2 = (t_byte) s[i];
	if ((b2 & 0x80) != 0x80) {
	  throw Exception ("decode-error", 
			   "invalid byte while reading utf-8 sequence");
	}
	buf[idx] |= ((t_quad) (b2 & 0x3F)) << 18;
	if (++i >= size) {
	  throw Exception ("decode-error", 
			   "invalid eos while reading utf-8 sequence");
	}
	t_byte b3 = (t_byte) s[i];
	if ((b3 & 0x80) != 0x80) {
	  throw Exception ("decode-error", 
			   "invalid byte while reading utf-8 sequence");
	}
	buf[idx] |= (t_quad) (b3 & 0x3F) << 12;
	if (++i >= size) {
	  throw Exception ("decode-error", 
			   "invalid eos while reading utf-8 sequence");
	}
	t_byte b4 = (t_byte) s[i];
	if ((b4 & 0x80) != 0x80) {
	  throw Exception ("decode-error", 
			   "invalid byte while reading utf-8 sequence");
	}
	buf[idx] |= (t_quad) (b4 & 0x3F) << 6;
	if (++i >= size) {
	  throw Exception ("decode-error", 
			   "invalid eos while reading utf-8 sequence");
	}
	t_byte b5 = (t_byte) s[i];
	if ((b5 & 0x80) != 0x80) {
	  throw Exception ("decode-error", 
			   "invalid byte while reading utf-8 sequence");
	}
	buf[idx] |= (t_quad) (b5 & 0x3F);
	if (buf[idx++] < 0x00200000) {
	  throw Exception ("decode-error", "invalid long utf-8 sequence");
	}
	continue;
      }
      // 6 bytes mode
      if (b1 < 0xFE) {
	buf[idx] = ((t_quad) (b1 & 0x01)) << 30;
	if (++i >= size) {
	  throw Exception ("decode-error", 
			   "invalid eos while reading utf-8 sequence");
	}
	t_byte b2 = (t_byte) s[i];
	if ((b2 & 0x80) != 0x80) {
	  throw Exception ("decode-error", 
			   "invalid byte while reading utf-8 sequence");
	}
	buf[idx] |= ((t_quad) (b2 & 0x3F)) << 24;
	if (++i >= size) {
	  throw Exception ("decode-error", 
			   "invalid eos while reading utf-8 sequence");
	}
	t_byte b3 = (t_byte) s[i];
	if ((b3 & 0x80) != 0x80) {
	  throw Exception ("decode-error", 
			   "invalid byte while reading utf-8 sequence");
	}
	buf[idx] |= (t_quad) (b3 & 0x3F) << 18;
	if (++i >= size) {
	  throw Exception ("decode-error", 
			   "invalid eos while reading utf-8 sequence");
	}
	t_byte b4 = (t_byte) s[i];
	if ((b4 & 0x80) != 0x80) {
	  throw Exception ("decode-error", 
			   "invalid byte while reading utf-8 sequence");
	}
	buf[idx] |= (t_quad) (b4 & 0x3F) << 12;
	if (++i >= size) {
	  throw Exception ("decode-error", 
			   "invalid eos while reading utf-8 sequence");
	}
	t_byte b5 = (t_byte) s[i];
	if ((b5 & 0x80) != 0x80) {
	  throw Exception ("decode-error", 
			   "invalid byte while reading utf-8 sequence");
	}
	buf[idx] |= (t_quad) (b5 & 0x3F) << 6;
	if (++i >= size) {
	  throw Exception ("decode-error", 
			   "invalid eos while reading utf-8 sequence");
	}
	t_byte b6 = (t_byte) s[i];
	if ((b6 & 0x80) != 0x80) {
	  throw Exception ("decode-error", 
			   "invalid byte while reading utf-8 sequence");
	}
	buf[idx] |= (t_quad) (b6 & 0x3F);
	if (buf[idx++] < 0x04000000) {
	  throw Exception ("decode-error", "invalid long utf-8 sequence");
	}
	continue;
      }
      throw Exception ("decode-error", "invalid utf-8 character sequence");
    }
    // add the nil quad
    buf[idx++] = nilq;
    // here we are
    return buf;
  }
}
