// ---------------------------------------------------------------------------
// - Cookie.cpp                                                              -
// - afnix:www module - http cookie class implementation                     -
// ---------------------------------------------------------------------------
// - This program is free software;  you can redistribute it  and/or  modify -
// - it provided that this copyright notice is kept intact.                  -
// -                                                                         -
// - This program  is  distributed in  the hope  that it will be useful, but -
// - without  any  warranty;  without  even   the   implied    warranty   of -
// - merchantability or fitness for a particular purpose.  In no event shall -
// - the copyright holder be liable for any  direct, indirect, incidental or -
// - special damages arising in any way out of the use of this software.     -
// ---------------------------------------------------------------------------
// - copyright (c) 1999-2006 amaury darsch                                   -
// ---------------------------------------------------------------------------

#include "Date.hpp"
#include "Vector.hpp"
#include "Cookie.hpp"
#include "Integer.hpp"
#include "Boolean.hpp"
#include "QuarkZone.hpp"
#include "Exception.hpp"

namespace afnix {

  // -------------------------------------------------------------------------
  // - private section                                                       -
  // -------------------------------------------------------------------------

  // format the date for a cookie expire time
  static String fmt_cookie (const t_long mage) {
    // prepare for the time object
    Date date;
    date.add (mage);
    // get the day of week
    String result = date.mapwday ();
    result = result + ", ";
    // get the day in month
    long mday = date.getmday ();
    if (mday < 10) {
      result = result + '0' + mday + '-';
    } else {
      result = result + mday + '-';
    }
    // get the year month
    result = result + date.mapymon () + '-';
    // get the year
    long year = date.getyear ();
    result = result + year + ' ';
    // get the time
    result = result + date.Time::format (true) + " GMT";
    return result;
  }

  // -------------------------------------------------------------------------
  // - class section                                                         -
  // -------------------------------------------------------------------------

  // create a cookie with a name/value pair

  Cookie::Cookie (const String& name, const String& value) {
    d_name   = name;
    d_value  = value;
    d_mage   = -1LL;
    d_secure = false;
  }

  // create a cookie with a name/value pair

  Cookie::Cookie (const String& name, const String& value, const t_long mage) {
    d_name   = name;
    d_value  = value;
    d_mage   = (mage < 0LL) ? -1LL : mage;
    d_secure = false;
  }

  // return the class name

  String Cookie::repr (void) const {
    return "Cookie";
  }

  // return a http cookie string representation

  String Cookie::tostring (void) const {
    rdlock ();
    if ((d_name.length () == 0) || (d_value.length () == 0)) {
      unlock ();
      throw Exception ("cookie-error", "invalid cookie name or value");
    }
    String result = "Set-Cookie: ";
    result = result + d_name.toliteral () + '=' + d_value.toliteral ();
    if (d_mage >= 0) {
      result = result + "; Max-Age=" + d_mage;
    }
    if (d_path.length () != 0) {
      result = result + "; Path=" + d_path.toliteral ();
    }
    if (d_domain.length () != 0) {
      result = result + "; Domain=" + d_domain.toliteral ();
    }
    if (d_comment.length () != 0) {
      result = result + "; Comment=" + d_comment.toliteral ();
    }
    if (d_mage > 0) {
      result = result + "; Expires=" + fmt_cookie (d_mage);
    }
    if (d_secure == true) {
      result = result + "; Secure";
    }
    unlock ();
    return result;
  }

  // set the cookie name
  
  void Cookie::setname (const String& name) {
    wrlock ();
    d_name = name;
    unlock ();
  }

  // get the cookie name

  String Cookie::getname (void) const {
    rdlock ();
    String result = d_name;
    unlock ();
    return result;
  }

  // set the cookie value
  
  void Cookie::setvalue (const String& value) {
    wrlock ();
    d_value = value;
    unlock ();
  }

  // get the cookie value

  String Cookie::getvalue (void) const {
    rdlock ();
    String result = d_value;
    unlock ();
    return result;
  }

  // set the cookie maximum age
  
  void Cookie::setmage (const t_long mage) {
    wrlock ();
    d_mage = (mage < 0LL) ? -1LL : mage;
    unlock ();
  }

  // get the cookie maximum age

  t_long Cookie::getmage (void) const {
    rdlock ();
    t_long result = d_mage;
    unlock ();
    return result;
  }

  // set the cookie path
  
  void Cookie::setpath (const String& path) {
    wrlock ();
    d_path = path;
    unlock ();
  }

  // get the cookie path

  String Cookie::getpath (void) const {
    rdlock ();
    String result = d_path;
    unlock ();
    return result;
  }

  // set the cookie domain
  
  void Cookie::setdomain (const String& domain) {
    wrlock ();
    d_domain = domain;
    unlock ();
  }

  // get the cookie domain

  String Cookie::getdomain (void) const {
    rdlock ();
    String result = d_domain;
    unlock ();
    return result;
  }

  // set the cookie comment
  
  void Cookie::setcomment (const String& comment) {
    wrlock ();
    d_comment = comment;
    unlock ();
  }

  // get the cookie comment

  String Cookie::getcomment (void) const {
    rdlock ();
    String result = d_comment;
    unlock ();
    return result;
  }

  // set the cookie secure flag
  
  void Cookie::setsecure (const bool flag) {
    wrlock ();
    d_secure = flag;
    unlock ();
  }

  // get the cookie secure flag

  bool Cookie::getsecure (void) const {
    rdlock ();
    bool result = d_secure;
    unlock ();
    return result;
  }
  
  // -------------------------------------------------------------------------
  // - object section                                                        -
  // -------------------------------------------------------------------------

  // the quark zone
  static const long QUARK_ZONE_LENGTH = 15;
  static QuarkZone  zone (QUARK_ZONE_LENGTH);

  // the object supported quarks
  static const long QUARK_GETNAME    = zone.intern ("get-name");
  static const long QUARK_SETNAME    = zone.intern ("set-name");
  static const long QUARK_GETMAGE    = zone.intern ("get-max-age");
  static const long QUARK_SETMAGE    = zone.intern ("set-max-age");
  static const long QUARK_GETPATH    = zone.intern ("get-path");
  static const long QUARK_SETPATH    = zone.intern ("set-path");
  static const long QUARK_TOSTRING   = zone.intern ("to-string");
  static const long QUARK_GETVALUE   = zone.intern ("get-value");
  static const long QUARK_SETVALUE   = zone.intern ("set-value");
  static const long QUARK_GETDOMAIN  = zone.intern ("get-domain");
  static const long QUARK_SETDOMAIN  = zone.intern ("set-domain");
  static const long QUARK_GETSECURE  = zone.intern ("get-secure");
  static const long QUARK_SETSECURE  = zone.intern ("set-secure");
  static const long QUARK_GETCOMMENT = zone.intern ("get-comment");
  static const long QUARK_SETCOMMENT = zone.intern ("set-comment");

  // create a new object in a generic way

  Object* Cookie::mknew (Vector* argv) {
    long argc = (argv == nilp) ? 0 : argv->length ();

    // check for 2 arguments
    if (argc == 2) {
      String name  = argv->getstring (0);
      String value = argv->getstring (1);
      return new Cookie (name, value);
    }
    // check for 3 arguments
    if (argc == 3) {
      String name  = argv->getstring (0);
      String value = argv->getstring (1);
      long   mage  = argv->getint    (2);
      return new Cookie (name, value, mage);
    }
    // invalid arguments
    throw Exception ("argument-error", "invalid arguments with cookie");
  }

  // return true if the given quark is defined

  bool Cookie::isquark (const long quark, const bool hflg) const {
    rdlock ();
    if (zone.exists (quark) == true) {
      unlock ();
      return true;
    }
    bool result = hflg ? Object::isquark (quark, hflg) : false;
    unlock ();
    return result;
  }

  // apply this object with a set of arguments and a quark

  Object* Cookie::apply (Runnable* robj, Nameset* nset, const long quark,
			 Vector* argv) {
    // get the number of arguments
    long argc = (argv == nilp) ? 0 : argv->length ();
    
    // dispatch 0 argument
    if (argc == 0) {
      if (quark == QUARK_GETNAME)    return new String  (getname    ());
      if (quark == QUARK_GETVALUE)   return new String  (getvalue   ());
      if (quark == QUARK_GETMAGE)    return new Integer (getmage    ());
      if (quark == QUARK_GETPATH)    return new String  (getpath    ());
      if (quark == QUARK_GETDOMAIN)  return new String  (getdomain  ());
      if (quark == QUARK_GETCOMMENT) return new String  (getcomment ());
      if (quark == QUARK_GETSECURE)  return new Boolean (getsecure  ());
    }

    // dispatch 1 argument
    if (argc == 1) {
      if (quark == QUARK_SETNAME) {
	String name = argv->getstring (0);
	setname (name);
	return nilp;
      }
      if (quark == QUARK_SETVALUE) {
	String value = argv->getstring (0);
	setvalue (value);
	return nilp;
      }
      if (quark == QUARK_SETMAGE) {
	long mage = argv->getint (0);
	setmage (mage);
	return nilp;
      }      
      if (quark == QUARK_SETPATH) {
	String path = argv->getstring (0);
	setpath (path);
	return nilp;
      }      
      if (quark == QUARK_SETDOMAIN) {
	String domain = argv->getstring (0);
	setdomain (domain);
	return nilp;
      }
      if (quark == QUARK_SETCOMMENT) {
	String comment = argv->getstring (0);
	setcomment (comment);
	return nilp;
      }
      if (quark == QUARK_SETSECURE) {
	bool flag = argv->getbool (0);
	setsecure (flag);
	return nilp;
      }
    }
    // call the object method
    return Object::apply (robj, nset, quark, argv);
  }
}
