// ---------------------------------------------------------------------------
// - HtmlElement.cpp                                                         -
// - afnix:www module - html element class implementation                    -
// ---------------------------------------------------------------------------
// - This program is free software;  you can redistribute it  and/or  modify -
// - it provided that this copyright notice is kept intact.                  -
// -                                                                         -
// - This program  is  distributed in  the hope  that it will be useful, but -
// - without  any  warranty;  without  even   the   implied    warranty   of -
// - merchantability or fitness for a particular purpose.  In no event shall -
// - the copyright holder be liable for any  direct, indirect, incidental or -
// - special damages arising in any way out of the use of this software.     -
// ---------------------------------------------------------------------------
// - copyright (c) 1999-2006 amaury darsch                                   -
// ---------------------------------------------------------------------------

#include "Integer.hpp"
#include "Boolean.hpp"
#include "Runnable.hpp"
#include "QuarkZone.hpp"
#include "Exception.hpp"
#include "HtmlElement.hpp"

namespace afnix {

  // -------------------------------------------------------------------------
  // - class section                                                         -
  // -------------------------------------------------------------------------

  // create an element by name

  HtmlElement::HtmlElement (const String& name) {
    d_name = name;
    d_tags.reset ();
  }

  // return the class name

  String HtmlElement::repr (void) const {
    return "HtmlElement";
  }

  // reset this element

  void HtmlElement::reset (void) {
    wrlock ();
    d_tags.reset ();
    unlock ();
  }

  // set the element name (use with caution)

  void HtmlElement::setname (const String& name) {
    wrlock ();
    d_name = name;
    unlock ();
  }

  // return the element name

  String HtmlElement::getname (void) const {
    rdlock ();
    String result = d_name;
    unlock ();
    return result;
  }

  // return the number of tags

  long HtmlElement::taglen (void) const {
    rdlock ();
    long result = d_tags.length ();
    unlock ();
    return result;
  }

  // return true if a tag exists

  bool HtmlElement::istag (const String& name) const {
    rdlock ();
    try {
      long tlen = taglen ();
      for (long i = 0; i < tlen; i++) {
	HtmlTag* tag = gettag (i);
	if (tag->getname () == name) {
	  unlock ();
	  return true;
	}
      }
      unlock ();
      return false;
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // add a tag to this element

  void HtmlElement::addtag (HtmlTag* tag) {
    if (tag == nilp) return;
    wrlock ();
    try {
      if (istag (tag->getname ()) == false) {
	d_tags.append (tag);
      }
      unlock ();
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // add a tag to this element by name and value

  void HtmlElement::addtag (const String& name, const Literal& value) {
    HtmlTag* tag = new HtmlTag (name, value);
    addtag (tag);
  }

  // get a tag by index

  HtmlTag* HtmlElement::gettag (const long index) const {
    rdlock ();
    try {
      HtmlTag* tag = dynamic_cast <HtmlTag*> (d_tags.get (index));
      if (tag == nilp) {
	throw Exception ("internal-error", "invalid object in tag vector");
      }
      unlock ();
      return tag;
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // return a opening string of the element
  
  String HtmlElement::opening (void) const {
    rdlock ();
    String result = '<';
    result = result + d_name;
    // add the tags
    long tlen = d_tags.length ();
    for (long i = 0; i < tlen; i++) {
      HtmlTag* tag = dynamic_cast <HtmlTag*> (d_tags.get (i));
      if (tag == nilp) continue;
      result = result + ' ' + tag->tostring ();
    }
    result = result + '>';
    unlock ();
    return result;
  }

  // return a closing string of the element
  
  String HtmlElement::closing (void) const {
    rdlock ();
    String result = "</";
    result = result + d_name + '>';
    unlock ();
    return result;
  }

  // emit this element into a buffer

  void HtmlElement::emit (Buffer& buf) const {
    rdlock  ();
    buf.add (opening ());
    buf.add (closing ());
    buf.add (eolc);
    unlock ();
  }

  // -------------------------------------------------------------------------
  // - object section                                                        -
  // -------------------------------------------------------------------------

  // the quark zone
  static const long QUARK_ZONE_LENGTH = 9;
  static QuarkZone  zone (QUARK_ZONE_LENGTH);

  // the object supported quarks
  static const long QUARK_EMIT    = zone.intern ("emit");
  static const long QUARK_RESET   = zone.intern ("reset");
  static const long QUARK_TAGP    = zone.intern ("tag-p");
  static const long QUARK_TAGLEN  = zone.intern ("tag-length");
  static const long QUARK_ADDTAG  = zone.intern ("add-tag");
  static const long QUARK_GETTAG  = zone.intern ("get-tag");
  static const long QUARK_GETNAME = zone.intern ("get-name");
  static const long QUARK_OPENING = zone.intern ("opening");
  static const long QUARK_CLOSING = zone.intern ("closing");

  // create a new oject in a generic way

  Object* HtmlElement::mknew (Vector* argv) {
    long argc = (argv == nilp) ? 0 : argv->length ();

    // check for 1 argument
    if (argc == 1) {
      String name = argv->getstring (0);
      return new HtmlElement (name);
    }

    // invalid arguments
    throw Exception ("argument-error", 
		     "invalid arguments with html element object");
  }

  // return true if the given quark is defined

  bool HtmlElement::isquark (const long quark, const bool hflg) const {
    rdlock ();
    if (zone.exists (quark) == true) {
      unlock ();
      return true;
    }
    bool result = hflg ? Object::isquark (quark, hflg) : false;
    unlock ();
    return result;
  }

  // apply this object with a set of arguments and a quark

  Object* HtmlElement::apply (Runnable* robj, Nameset* nset, const long quark,
			      Vector* argv) {
    // get the number of arguments
    long argc = (argv == nilp) ? 0 : argv->length ();

    // dispatch 0 argument
    if (argc == 0) {
      if (quark == QUARK_TAGLEN)  return new Integer (taglen ());
      if (quark == QUARK_GETNAME) return new String  (getname ());
      if (quark == QUARK_OPENING) return new String  (opening ());
      if (quark == QUARK_CLOSING) return new String  (closing ());
      if (quark == QUARK_RESET) {
	reset ();
	return nilp;
      }
    }

    // dispatch 1 argument
    if (argc == 1) {
      if (quark == QUARK_EMIT) {
	Buffer* buf = dynamic_cast <Buffer*> (argv->get (0));
	if (buf == nilp) {
	  throw Exception ("type-error", "buffer expected with emit");
	}
	emit (*buf);
	return nilp;
      }
      if (quark == QUARK_TAGP) {
	String name = argv->getstring (0);
	return new Boolean (istag (name));
      }
      if (quark == QUARK_GETTAG) {
	long index = argv->getint (0);
	try {
	  rdlock ();
	  HtmlTag* result = gettag (index);
	  robj->post (result);
	  unlock ();
	  return result;
	} catch (...) {
	  unlock ();
	  throw;
	}
      }
      if (quark == QUARK_ADDTAG) {
	HtmlTag* tag = dynamic_cast <HtmlTag*> (argv->get (0));
	if (tag == nilp) {
	  throw Exception ("type-error", "tag expected with add-tag");
	}
	addtag (tag);
	return nilp;
      }
    }

    // dispatch 2 arguments
    if (argc == 2) {
      if (quark == QUARK_ADDTAG) {
	String   name = argv->getstring (0);
	Literal* tval = dynamic_cast <Literal*> (argv->get (1));
	if (tval == nilp) {
	  throw Exception ("type-error", "literal expected with add-tag");
	}
	addtag (name, *tval);
	return nilp;
      }
    }
    // call the object method
    return Object::apply (robj, nset, quark, argv);
  }
}
