// ---------------------------------------------------------------------------
// - HtmlTable.cpp                                                           -
// - afnix:www module - html table class implementation                      -
// ---------------------------------------------------------------------------
// - This program is free software;  you can redistribute it  and/or  modify -
// - it provided that this copyright notice is kept intact.                  -
// -                                                                         -
// - This program  is  distributed in  the hope  that it will be useful, but -
// - without  any  warranty;  without  even   the   implied    warranty   of -
// - merchantability or fitness for a particular purpose.  In no event shall -
// - the copyright holder be liable for any  direct, indirect, incidental or -
// - special damages arising in any way out of the use of this software.     -
// ---------------------------------------------------------------------------
// - copyright (c) 1999-2006 amaury darsch                                   -
// ---------------------------------------------------------------------------

#include "Vector.hpp"
#include "Integer.hpp"
#include "Runnable.hpp"
#include "HtmlTable.hpp"
#include "QuarkZone.hpp"
#include "Exception.hpp"

namespace afnix {

  // -------------------------------------------------------------------------
  // - class section                                                         -
  // -------------------------------------------------------------------------

  // create a new html table

  HtmlTable::HtmlTable (void) : HtmlElement ("table") {
    d_thrs.reset ();
    d_tbrs.reset ();
    d_tfrs.reset ();
  }

  // create a new html table with a default tr class

  HtmlTable::HtmlTable (const String& dtrc) : HtmlElement ("table") {
    d_htrc = dtrc;
    d_btrc = dtrc;
    d_ftrc = dtrc;
    d_thrs.reset ();
    d_tbrs.reset ();
    d_tfrs.reset ();
  }

  // create a new html table with a default tr and td class

  HtmlTable::HtmlTable (const String& dtrc, const String& dtdc) : 
    HtmlElement ("table") {
    d_htrc = dtrc;
    d_btrc = dtrc;
    d_ftrc = dtrc;
    d_htdc = dtdc;
    d_btdc = dtdc;
    d_ftdc = dtdc;
    d_thrs.reset ();
    d_tbrs.reset ();
    d_tfrs.reset ();
  }

  // create a new html table with a default class and tr and td class

  HtmlTable::HtmlTable (const String& dtbc, const String& dtrc, 
			const String& dtdc) : HtmlElement ("table") {
    // save the default classes
    d_htrc = dtrc;
    d_btrc = dtrc;
    d_ftrc = dtrc;
    d_htdc = dtdc;
    d_btdc = dtdc;
    d_ftdc = dtdc;
    // reset the tables
    d_thrs.reset ();
    d_tbrs.reset ();
    d_tfrs.reset ();
    // set the table class
    addtag ("class", dtbc);
  }

  // create a new html table with a print table

  HtmlTable::HtmlTable (const PrintTable& ptbl) : HtmlElement ("table") {
    d_thrs.reset ();
    d_tbrs.reset ();
    d_tfrs.reset ();
    addtbl (ptbl);
  }

  // return the class name

  String HtmlTable::repr (void) const {
    return "HtmlTable";
  }

  // set the default head tr class
  
  void HtmlTable::sethtrc (const String& dtrc) {
    wrlock ();
    try {
      // save the default tr class
      d_htrc = dtrc;
      // update all existing entries
      long hlen = gethlen ();
      for (long i = 0; i < hlen; i++) {
	HtmlTr* tr = gethtr (i);
	tr->addtag ("class", d_htrc);
      }
      unlock ();
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // get the default head tr class

  String HtmlTable::gethtrc (void) const {
    rdlock ();
    String result = d_htrc;
    unlock ();
    return result;
  }

  // set the default body tr class
  
  void HtmlTable::setbtrc (const String& dtrc) {
    wrlock ();
    try {
      // save the default tr class
      d_btrc = dtrc;
      // update all existing entries
      long blen = getblen ();
      for (long i = 0; i < blen; i++) {
	HtmlTr* tr = getbtr (i);
	tr->addtag ("class", d_btrc);
      }
      unlock ();
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // get the default body tr class

  String HtmlTable::getbtrc (void) const {
    rdlock ();
    String result = d_btrc;
    unlock ();
    return result;
  }

  // set the default footer tr class
  
  void HtmlTable::setftrc (const String& dtrc) {
    wrlock ();
    try {
      // save the default tr class
      d_ftrc = dtrc;
      // update all existing entries
      long flen = getflen ();
      for (long i = 0; i < flen; i++) {
	HtmlTr* tr = getftr (i);
	tr->addtag ("class", d_ftrc);
      }
      unlock ();
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // get the default body tr class

  String HtmlTable::getftrc (void) const {
    rdlock ();
    String result = d_ftrc;
    unlock ();
    return result;
  }

  // set the default head td class
  
  void HtmlTable::sethtdc (const String& dtdc) {
    wrlock ();
    try {
      // save the default td class
      d_htdc = dtdc;
      // update all existing entries
      long hlen = gethlen ();
      for (long i = 0; i < hlen; i++) {
	HtmlTr* tr = gethtr (i);
	tr->setdtdc (d_htdc);
      }
      unlock ();
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // get the default head td class

  String HtmlTable::gethtdc (void) const {
    rdlock ();
    String result = d_htdc;
    unlock ();
    return result;
  }

  // set the default body td class
  
  void HtmlTable::setbtdc (const String& dtdc) {
    wrlock ();
    try {
      // save the default td class
      d_btdc = dtdc;
      // update all existing entries
      long blen = getblen ();
      for (long i = 0; i < blen; i++) {
	HtmlTr* tr = getbtr (i);
	tr->setdtdc (d_btdc);
      }
      unlock ();
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // get the default body td class

  String HtmlTable::getbtdc (void) const {
    rdlock ();
    String result = d_btdc;
    unlock ();
    return result;
  }

  // set the default footer td class
  
  void HtmlTable::setftdc (const String& dtdc) {
    wrlock ();
    try {
      // save the default td class
      d_ftdc = dtdc;
      // update all existing entries
      long flen = getflen ();
      for (long i = 0; i < flen; i++) {
	HtmlTr* tr = getftr (i);
	tr->setdtdc (d_ftdc);
      }
      unlock ();
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // get the default footer td class

  String HtmlTable::getftdc (void) const {
    rdlock ();
    String result = d_ftdc;
    unlock ();
    return result;
  }

  // set the table caption

  void HtmlTable::setcaption (const String& caption) {
    wrlock ();
    d_caption = caption;
    unlock ();
  }

  // return the table caption

  String HtmlTable::getcaption (void) const {
    rdlock ();
    String result = d_caption;
    unlock ();
    return result;
  }

  // return the number of head rows

  long HtmlTable::gethlen (void) const {
    rdlock ();
    long result = d_thrs.length ();
    unlock ();
    return result;
  }

  // return the number of body rows

  long HtmlTable::getblen (void) const {
    rdlock ();
    long result = d_tbrs.length ();
    unlock ();
    return result;
  }

  // return the number of footer rows

  long HtmlTable::getflen (void) const {
    rdlock ();
    long result = d_tfrs.length ();
    unlock ();
    return result;
  }

  // get a head row element by index

  HtmlTr* HtmlTable::gethtr (const long index) const {
    rdlock ();
    try {
      HtmlTr* tr = dynamic_cast <HtmlTr*> (d_thrs.get (index));
      if (tr == nilp) {
	throw Exception ("internal-error", "invalid object in tr vector");
      }
      unlock ();
      return tr;
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // get a body row element by index

  HtmlTr* HtmlTable::getbtr (const long index) const {
    rdlock ();
    try {
      HtmlTr* tr = dynamic_cast <HtmlTr*> (d_tbrs.get (index));
      if (tr == nilp) {
	throw Exception ("internal-error", "invalid object in tr vector");
      }
      unlock ();
      return tr;
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // get a footer row element by index

  HtmlTr* HtmlTable::getftr (const long index) const {
    rdlock ();
    try {
      HtmlTr* tr = dynamic_cast <HtmlTr*> (d_tfrs.get (index));
      if (tr == nilp) {
	throw Exception ("internal-error", "invalid object in tr vector");
      }
      unlock ();
      return tr;
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // add a row to the head table

  void HtmlTable::addhead (HtmlTr* tr) {
    if (tr == nilp) return;
    wrlock ();
    try {
      // check for the default class
      if (d_htrc.isnil () == false) tr->addtag  ("class", d_htrc);
      if (d_htdc.isnil () == false) tr->setdtdc (d_htdc);
      // add the head table row
      d_thrs.append (tr);
      unlock ();
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // add a row to the body table

  void HtmlTable::addbody (HtmlTr* tr) {
    if (tr == nilp) return;
    wrlock ();
    try {
      // check for the default class
      if (d_btrc.isnil () == false) tr->addtag ("class", d_btrc);
      if (d_btdc.isnil () == false) tr->setdtdc (d_btdc);
      // add the body table row
      d_tbrs.append (tr);
      unlock ();
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // add a row to the footer table

  void HtmlTable::addfoot (HtmlTr* tr) {
    if (tr == nilp) return;
    wrlock ();
    try {
      // check for the default class
      if (d_ftrc.isnil () == false) tr->addtag ("class", d_ftrc);
      if (d_ftdc.isnil () == false) tr->setdtdc (d_ftdc);
      // add the footer table row
      d_tfrs.append (tr);
      unlock ();
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // add a print table to the table

  void HtmlTable::addtbl (const PrintTable& ptbl) {
    // get the number of rows and columns
    long rows = ptbl.getrows ();
    long cols = ptbl.getcols ();
    if ((rows == 0) || (cols == 0)) return;
    wrlock ();
    try {
      for (long i = 0; i < rows; i++) {
	HtmlTr* tr = new HtmlTr (d_btdc);
	for (long j = 0; j < cols; j++) {
	  // get the data
	  String data = ptbl.get (i, j);
	  HtmlTableData* td = tr->adddata (data);
	  // get the direction
	  if (ptbl.getcdir (j) == true) {
	    td->addtag ("align", String("right"));
	  }
	}
	addbody (tr);
      }
    } catch (...) {
      unlock ();
      throw;
    }
    unlock ();
  }

  // emit this table row into a buffer

  void HtmlTable::emit (Buffer& buf) const {
    rdlock  ();
    // open the table
    buf.add (opening ()); buf.add (eolc);
    // emit the caption
    if (d_caption.length () > 0) {
      buf.add (String ("<caption>"));
      buf.add (d_caption);
      buf.add (String ("</caption>\n"));
    }
    // emit the head
    long hlen = d_thrs.length ();
    if (hlen > 0) {
      buf.add (String ("<thead>\n"));
      for (long i = 0; i < hlen; i++) {
	HtmlTr* tr = dynamic_cast <HtmlTr*> (d_thrs.get (i));
	if (tr != nilp) tr->emit (buf);
      }
      buf.add (String ("</thead>\n"));
    }
    // emit the body
    long blen = d_tbrs.length ();
    if (blen > 0) {
      buf.add (String ("<tbody>\n"));
      for (long i = 0; i < blen; i++) {
	HtmlTr* tr = dynamic_cast <HtmlTr*> (d_tbrs.get (i));
	if (tr != nilp) tr->emit (buf);
      }
      buf.add (String ("</tbody>\n"));
    }
    // emit the footer
    long flen = d_tfrs.length ();
    if (flen > 0) {
      buf.add (String ("<tfoot>\n"));
      for (long i = 0; i < flen; i++) {
	HtmlTr* tr = dynamic_cast <HtmlTr*> (d_tfrs.get (i));
	if (tr != nilp) tr->emit (buf);
      }
      buf.add (String ("</tfoot>\n"));
    }
    // close the table
    buf.add (closing ()); buf.add (eolc);
    unlock ();
  }

  // -------------------------------------------------------------------------
  // - object section                                                        -
  // -------------------------------------------------------------------------

  // the quark zone
  static const long QUARK_ZONE_LENGTH = 24;
  static QuarkZone  zone (QUARK_ZONE_LENGTH);

  // the object supported quarks
  static const long QUARK_GETHTR     = String::intern ("get-head-tr");
  static const long QUARK_GETBTR     = String::intern ("get-body-tr");
  static const long QUARK_GETFTR     = String::intern ("get-foot-tr");
  static const long QUARK_SETHTRC    = String::intern ("set-head-tr-class");
  static const long QUARK_GETHTRC    = String::intern ("get-head-tr-class");
  static const long QUARK_SETBTRC    = String::intern ("set-body-tr-class");
  static const long QUARK_GETBTRC    = String::intern ("get-body-tr-class");
  static const long QUARK_SETFTRC    = String::intern ("set-foot-tr-class");
  static const long QUARK_GETFTRC    = String::intern ("get-foot-tr-class");
  static const long QUARK_SETHTDC    = String::intern ("set-head-td-class");
  static const long QUARK_GETHTDC    = String::intern ("get-head-td-class");
  static const long QUARK_SETBTDC    = String::intern ("set-body-td-class");
  static const long QUARK_GETBTDC    = String::intern ("get-body-td-class");
  static const long QUARK_SETFTDC    = String::intern ("set-foot-td-class");
  static const long QUARK_GETFTDC    = String::intern ("get-foot-td-class");
  static const long QUARK_ADDHEAD    = String::intern ("add-head");
  static const long QUARK_ADDBODY    = String::intern ("add-body");
  static const long QUARK_ADDFOOT    = String::intern ("add-foot");
  static const long QUARK_ADDTABLE   = String::intern ("add-table");
  static const long QUARK_HEADLENGTH = String::intern ("head-length");
  static const long QUARK_BODYLENGTH = String::intern ("body-length");
  static const long QUARK_FOOTLENGTH = String::intern ("foot-length");
  static const long QUARK_GETCAPTION = String::intern ("get-caption");
  static const long QUARK_SETCAPTION = String::intern ("set-caption");

  // create a object in a generic way

  Object* HtmlTable::mknew (Vector* argv) {
    long argc = (argv == nilp) ? 0 : argv->length ();
    // check for 0 argument
    if (argc == 0) return new HtmlTable;
    // check for 1 argument
    if (argc == 1) {
      Object* obj = argv->get (0);
      // check for a string
      String* sobj = dynamic_cast <String*> (obj);
      if (sobj != nilp) return new HtmlTable (*sobj);
      // check for a print table
      PrintTable* ptbl = dynamic_cast <PrintTable*> (obj);
      if (ptbl != nilp) return new HtmlTable (*ptbl);
      // invalid argument
      throw Exception ("type-error", "invalid argmument with html table",
		       Object::repr (obj));
    }
    // check for 2 arguments
    if (argc == 2) {
      String dtrc = argv->getstring (0);
      String dtdc = argv->getstring (1);
      return new HtmlTable (dtrc, dtdc);
    }
    // check for 3 arguments
    if (argc == 3) {
      String dtbc = argv->getstring (0);
      String dtrc = argv->getstring (1);
      String dtdc = argv->getstring (2);
      return new HtmlTable (dtbc, dtrc, dtdc);
    }
    // invalid arguments
    throw Exception ("argument-error", 
		     "too many arguments with html table object");
  }

  // return true if the given quark is defined

  bool HtmlTable::isquark (const long quark, const bool hflg) const {
    rdlock ();
    if (zone.exists (quark) == true) {
      unlock ();
      return true;
    }
    bool result = hflg ? Object::isquark (quark, hflg) : false;
    unlock ();
    return result;
  }
  
  // apply this object with a set of arguments and a quark

  Object* HtmlTable::apply (Runnable* robj, Nameset* nset, const long quark,
			    Vector* argv) {
    // get the number of arguments
    long argc = (argv == nilp) ? 0 : argv->length ();
 
    // dispatch 0 argument
    if (argc == 0) {
      if (quark == QUARK_GETHTRC)    return new String  (gethtrc ());
      if (quark == QUARK_GETBTRC)    return new String  (getbtrc ());
      if (quark == QUARK_GETFTRC)    return new String  (getftrc ());
      if (quark == QUARK_GETHTDC)    return new String  (gethtdc ());
      if (quark == QUARK_GETBTDC)    return new String  (getbtdc ());
      if (quark == QUARK_GETFTDC)    return new String  (getftdc ());
      if (quark == QUARK_HEADLENGTH) return new Integer (gethlen ());
      if (quark == QUARK_BODYLENGTH) return new Integer (getblen ());
      if (quark == QUARK_FOOTLENGTH) return new Integer (getflen ());
      if (quark == QUARK_GETCAPTION) return new String  (getcaption ());
    }
	
    // dispatch 1 argument
    if (argc == 1) {
      if (quark == QUARK_GETHTR) {
	long index = argv->getint (0);
	rdlock ();
	try {
	  HtmlTr* result = gethtr (index);
	  robj->post (result);
	  unlock ();
	  return result;
	} catch (...) {
	  unlock ();
	  throw;
	}
      }
      if (quark == QUARK_GETBTR) {
	long index = argv->getint (0);
	rdlock ();
	try {
	  HtmlTr* result = getbtr (index);
	  robj->post (result);
	  unlock ();
	  return result;
	} catch (...) {
	  unlock ();
	  throw;
	}
      }
      if (quark == QUARK_GETFTR) {
	long index = argv->getint (0);
	rdlock ();
	try {
	  HtmlTr* result = getftr (index);
	  robj->post (result);
	  unlock ();
	  return result;
	} catch (...) {
	  unlock ();
	  throw;
	}
      }
      if (quark == QUARK_SETHTRC) {
	String dtrc = argv->getstring (0);
	sethtrc (dtrc);
	return nilp;
      }
      if (quark == QUARK_SETBTRC) {
	String dtrc = argv->getstring (0);
	setbtrc (dtrc);
	return nilp;
      }
      if (quark == QUARK_SETFTRC) {
	String dtrc = argv->getstring (0);
	setftrc (dtrc);
	return nilp;
      }
      if (quark == QUARK_SETHTDC) {
	String dtdc = argv->getstring (0);
	sethtdc (dtdc);
	return nilp;
      }
      if (quark == QUARK_SETBTDC) {
	String dtdc = argv->getstring (0);
	setbtdc (dtdc);
	return nilp;
      }
      if (quark == QUARK_SETFTDC) {
	String dtdc = argv->getstring (0);
	setftdc (dtdc);
	return nilp;
      }
      if (quark == QUARK_SETCAPTION) {
	String caption = argv->getstring (0);
	setcaption (caption);
	return nilp;
      }
      if (quark == QUARK_ADDHEAD) {
	HtmlTr* tr = dynamic_cast <HtmlTr*> (argv->get (0));
	if (tr == nilp)
	  throw Exception ("type-error", "row element expected with add-head");
	addhead (tr);
	return nilp;
      }
      if (quark == QUARK_ADDBODY) {
	HtmlTr* tr = dynamic_cast <HtmlTr*> (argv->get (0));
	if (tr == nilp)
	  throw Exception ("type-error", "row element expected with add-body");
	addbody (tr);
	return nilp;
      }
      if (quark == QUARK_ADDFOOT) {
	HtmlTr* tr = dynamic_cast <HtmlTr*> (argv->get (0));
	if (tr == nilp)
	  throw Exception ("type-error", "row element expected with add-foot");
	addfoot (tr);
	return nilp;
      }
      if (quark == QUARK_ADDTABLE) {
	PrintTable* tbl = dynamic_cast <PrintTable*> (argv->get (0));
	if (tbl == nilp)
	  throw Exception ("type-error", "print table expected for add-table");
	addtbl (*tbl);
	return nilp;
      }
    }
    // call the html element method
    return HtmlElement::apply (robj, nset, quark, argv);
  }
}
