// ---------------------------------------------------------------------------
// - HtmlTd.cpp                                                              -
// - afnix:www module - html table data class implementation                 -
// ---------------------------------------------------------------------------
// - This program is free software;  you can redistribute it  and/or  modify -
// - it provided that this copyright notice is kept intact.                  -
// -                                                                         -
// - This program  is  distributed in  the hope  that it will be useful, but -
// - without  any  warranty;  without  even   the   implied    warranty   of -
// - merchantability or fitness for a particular purpose.  In no event shall -
// - the copyright holder be liable for any  direct, indirect, incidental or -
// - special damages arising in any way out of the use of this software.     -
// ---------------------------------------------------------------------------
// - copyright (c) 1999-2006 amaury darsch                                   -
// ---------------------------------------------------------------------------

#include "Vector.hpp"
#include "HtmlTd.hpp"
#include "Integer.hpp"
#include "Runnable.hpp"
#include "QuarkZone.hpp"
#include "Exception.hpp"

namespace afnix {

  // -------------------------------------------------------------------------
  // - class section (table date)                                            -
  // -------------------------------------------------------------------------

  // create a default table element by name

  HtmlTableData::HtmlTableData (const String& name) : HtmlElement (name) {
    d_data = "&nbsp;";
  }

  // set the table data element by literal

  void HtmlTableData::setdata (const Literal& lval) {
    wrlock ();
    d_data = lval.tostring ();
    if (d_data.length () == 0) d_data = "&nbsp;";
    unlock ();
  }

  // return the table data element

  String HtmlTableData::getdata (void) const{
    rdlock ();
    String result = d_data;
    unlock ();
    return result;
  }

  // emit this table data into a buffer

  void HtmlTableData::emit (Buffer& buf) const {
    rdlock  ();
    buf.add (opening ());
    buf.add (d_data);
    buf.add (closing ()); 
    buf.add (eolc);
    unlock ();
  }

  // -------------------------------------------------------------------------
  // - object section                                                        -
  // -------------------------------------------------------------------------

  // the quark zone
  static const long QUARK_ZONE_LENGTH = 2;
  static QuarkZone  zone (QUARK_ZONE_LENGTH);

  // the object supported quarks
  static const long QUARK_SETDATA  = zone.intern ("set-data");
  static const long QUARK_GETDATA  = zone.intern ("get-data");

  // return true if the given quark is defined

  bool HtmlTableData::isquark (const long quark, const bool hflg) const {
    rdlock ();
    if (zone.exists (quark) == true) {
      unlock ();
      return true;
    }
    bool result = hflg ? Object::isquark (quark, hflg) : false;
    unlock ();
    return result;
  }

  // apply this object with a set of arguments and a quark

  Object* HtmlTableData::apply (Runnable* robj, Nameset* nset, 
				const long quark, Vector* argv) {
    // get the number of arguments
    long argc = (argv == nilp) ? 0 : argv->length ();
    
    // dispatch 0 argument
    if (argc == 0) {
      if (quark == QUARK_GETDATA) return new String (getdata ());
    }
    // dispatch 1 argument
    if (argc == 1) {
      if (quark == QUARK_SETDATA) {
	Literal* lval = dynamic_cast <Literal*> (argv->get (0));
	if (lval == nilp)
	  throw Exception ("type-error", "literal expected with set-data");
	setdata (*lval);
	return nilp;
      }
    }
    // call the html element method
    return HtmlElement::apply (robj, nset, quark, argv);
  }

  // -------------------------------------------------------------------------
  // - class section  (td element)                                           -
  // -------------------------------------------------------------------------
  
  // create an empty table data

  HtmlTd::HtmlTd (void) : HtmlTableData ("td") {
    d_data = "&nbsp;";
  }

  // create a table data element by value

  HtmlTd::HtmlTd (const Literal& lval) : HtmlTableData ("td") {
    d_data = lval.tostring ();
    if (d_data.length () == 0) d_data = "&nbsp;";
  }


  // create a table data element by value and class
  HtmlTd::HtmlTd (const Literal& lval, const String& dtdc) : 
          HtmlTableData ("td") {
    // set the td data
    d_data = lval.tostring ();
    if (d_data.isnil () == true) d_data = "&nbsp;";
    // set the default class
    if (dtdc.isnil () == false) {
      HtmlTag* tag = new HtmlTag ("class", dtdc);
      addtag (tag);
    }
  }

  // return the class name

  String HtmlTd::repr (void) const {
    return "HtmlTd";
  }

  // create a new object in a generic way

  Object* HtmlTd::mknew (Vector* argv) {
    long argc = (argv == nilp) ? 0 : argv->length ();
    // check for 0 argument
    if (argc == 0) return new HtmlTd;
    // check for 1 argument
    if (argc == 1) {
      Literal* lval = dynamic_cast <Literal*> (argv->get (0));
      if (lval == nilp)
	throw Exception ("argument-error",
			 "invalid argument with td element object");
      return new HtmlTd (*lval);
    }
    // check for 2 arguments
    if (argc == 2) {
      Literal* lval = dynamic_cast <Literal*> (argv->get (0));
      if (lval == nilp)
	throw Exception ("argument-error",
			 "invalid argument with td element object");
      String dtdc = argv->getstring (1);
      return new HtmlTd (*lval, dtdc);
    }
    // invalid arguments
    throw Exception ("argument-error", 
		     "too many arguments with html td element");
  }

  // -------------------------------------------------------------------------
  // - class section (th element)                                            -
  // -------------------------------------------------------------------------
  
  // create an empty table data

  HtmlTh::HtmlTh (void) : HtmlTableData ("th") {
    d_data = "&nbsp;";
  }

  // create a table data element by value

  HtmlTh::HtmlTh (const Literal& lval) : HtmlTableData ("th") {
    d_data = lval.tostring ();
    if (d_data.length () == 0) d_data = "&nbsp;";
  }


  // create a table data element by value and class
  HtmlTh::HtmlTh (const Literal& lval, const String& dtdc) : 
          HtmlTableData ("th") {
    // set the td data
    d_data = lval.tostring ();
    if (d_data.isnil () == true) d_data = "&nbsp;";
    // set the default class
    if (dtdc.isnil () == false) {
      HtmlTag* tag = new HtmlTag ("class", dtdc);
      addtag (tag);
    }
  }

  // return the class name

  String HtmlTh::repr (void) const {
    return "HtmlTh";
  }

  // create a object in a generic way

  Object* HtmlTh::mknew (Vector* argv) {
    long argc = (argv == nilp) ? 0 : argv->length ();
    // check for 0 argument
    if (argc == 0) return new HtmlTh;
    // check for 1 argument
    if (argc == 1) {
      Literal* lval = dynamic_cast <Literal*> (argv->get (0));
      if (lval == nilp)
	throw Exception ("argument-error",
			 "invalid argument with th element object");
      return new HtmlTh (*lval);
    }
    // check for 2 arguments
    if (argc == 2) {
      Literal* lval = dynamic_cast <Literal*> (argv->get (0));
      if (lval == nilp)
	throw Exception ("argument-error",
			 "invalid argument with th element object");
      String dtdc = argv->getstring (1);
      return new HtmlTh (*lval, dtdc);
    }
    // invalid arguments
    throw Exception ("argument-error", 
		     "too many arguments with html th element");
  }
}
