// ---------------------------------------------------------------------------
// - HtmlTr.cpp                                                              -
// - afnix:www module - html table row class implementation                  -
// ---------------------------------------------------------------------------
// - This program is free software;  you can redistribute it  and/or  modify -
// - it provided that this copyright notice is kept intact.                  -
// -                                                                         -
// - This program  is  distributed in  the hope  that it will be useful, but -
// - without  any  warranty;  without  even   the   implied    warranty   of -
// - merchantability or fitness for a particular purpose.  In no event shall -
// - the copyright holder be liable for any  direct, indirect, incidental or -
// - special damages arising in any way out of the use of this software.     -
// ---------------------------------------------------------------------------
// - copyright (c) 1999-2006 amaury darsch                                   -
// ---------------------------------------------------------------------------

#include "Vector.hpp"
#include "HtmlTr.hpp"
#include "Integer.hpp"
#include "Runnable.hpp"
#include "QuarkZone.hpp"
#include "Exception.hpp"

namespace afnix {

  // -------------------------------------------------------------------------
  // - class section                                                         -
  // -------------------------------------------------------------------------

  // create an empty table row

  HtmlTr::HtmlTr (void) : HtmlElement ("tr") {
    d_tbld.reset ();
  }

  // create a table row with a default class

  HtmlTr::HtmlTr (const String& dtdc) : HtmlElement ("tr") {
    d_dtdc = dtdc;
    d_tbld.reset ();
  }

  // return the class name

  String HtmlTr::repr (void) const {
    return "HtmlTr";
  }

  // return the number of data elements

  long HtmlTr::length (void) const {
    rdlock ();
    long result = d_tbld.length ();
    unlock ();
    return result;
  }

  // get a data element by index

  HtmlTableData* HtmlTr::gettd (const long index) const {
    rdlock ();
    try {
      HtmlTableData* td = dynamic_cast <HtmlTableData*> (d_tbld.get (index));
      if (td == nilp) {
	throw Exception ("internal-error", "invalid object in table data");
      }
      unlock ();
      return td;
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // set the default td class
  
  void HtmlTr::setdtdc (const String& dtdc) {
    wrlock ();
    try {
      // save the default td class
      d_dtdc = dtdc;
      // update all existing entries
      long tdlen = length ();for (long i = 0; i < tdlen; i++) {
	HtmlTableData* td = gettd (i);
	td->addtag ("class", d_dtdc);
      }
      unlock ();
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // get the default td class

  String HtmlTr::getdtdc (void) const {
    rdlock ();
    String result = d_dtdc;
    unlock ();
    return result;
  }
  
  // add a table data element to the row

  HtmlTableData* HtmlTr::adddata (HtmlTableData* td) {
    // check for nil td and lock
    if (td == nilp) return nilp;
    wrlock ();
    try {
      // check for the default class
      if (d_dtdc.isnil () == false) td->addtag ("class", d_dtdc);
      // add the table data
      d_tbld.append (td);
      unlock ();
      return td;
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // add a data literal to the row

  HtmlTableData* HtmlTr::adddata (const Literal& lval) {
    wrlock ();
    try {
      HtmlTableData* td = new HtmlTd (lval, d_dtdc);
      d_tbld.append (td);
      unlock ();
      return td;
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // add a header literal to the row

  HtmlTableData* HtmlTr::addhead (const Literal& lval) {
    wrlock ();
    try {
      HtmlTableData* th = new HtmlTh (lval, d_dtdc);
      d_tbld.append (th);
      unlock ();
      return th;
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // emit this table row into a buffer

  void HtmlTr::emit (Buffer& buf) const {
    rdlock  ();
    try {
      buf.add (opening ()); buf.add (eolc);
      long tlen = d_tbld.length ();
      for (long i = 0; i < tlen; i++) {
	HtmlTableData* td = dynamic_cast <HtmlTableData*> (d_tbld.get (i));
	if (td != nilp) td->emit (buf);
      }
      buf.add (closing ()); buf.add (eolc);
      unlock ();
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // -------------------------------------------------------------------------
  // - object section                                                        -
  // -------------------------------------------------------------------------

  // the quark zone
  static const long QUARK_ZONE_LENGTH = 6;
  static QuarkZone  zone (QUARK_ZONE_LENGTH);

  // the table data supported quarks
  static const long QUARK_GETTD    = zone.intern ("get-td");
  static const long QUARK_LENGTH   = zone.intern ("length");
  static const long QUARK_ADDDATA  = zone.intern ("add-data");
  static const long QUARK_ADDHEAD  = zone.intern ("add-head");
  static const long QUARK_SETDTDC  = zone.intern ("set-default-td-class");
  static const long QUARK_GETDTDC  = zone.intern ("get-default-td-class");

  // create a object in a generic way

  Object* HtmlTr::mknew (Vector* argv) {
    long argc = (argv == nilp) ? 0 : argv->length ();
    // check for 0 argument
    if (argc == 0) return new HtmlTr;
    // check for 1 argument
    if (argc == 1) {
      String dtdc = argv->getstring (0);
      return new HtmlTr (dtdc);
    }
    // invalid arguments
    throw Exception ("argument-error", 
		     "too many arguments with html table row object");
  }

  // return true if the given quark is defined

  bool HtmlTr::isquark (const long quark, const bool hflg) const {
    rdlock ();
    if (zone.exists (quark) == true) {
      unlock ();
      return true;
    }
    bool result = hflg ? Object::isquark (quark, hflg) : false;
    unlock ();
    return result;
  }

  // apply this object with a set of arguments and a quark

  Object* HtmlTr::apply (Runnable* robj, Nameset* nset, const long quark, 
			 Vector* argv) {
    // get the number of arguments
    long argc = (argv == nilp) ? 0 : argv->length ();

    // dispatch 0 argument
    if (argc == 0) {
      if (quark == QUARK_LENGTH)  return new Integer (length  ());
      if (quark == QUARK_GETDTDC) return new String  (getdtdc ());
    }

    // dispatch 1 argument
    if (argc == 1) {
      if (quark == QUARK_ADDDATA) {
	wrlock ();
	try {
	  Object* obj = argv->get (0);
	  HtmlTd* td = dynamic_cast <HtmlTd*> (obj);
	  if (td != nilp) {
	    HtmlTableData* result = adddata (td);
	    robj->post (result);
	    unlock ();
	    return result;
	  }
	  Literal* lval = dynamic_cast <Literal*> (obj);
	  if (lval != nilp) {
	    HtmlTableData* result = adddata (*lval);
	    robj->post (result);
	    unlock ();
	    return result;

	  }
	  throw Exception ("type-error", "unexpected object with add-data",
			   Object::repr (obj));
	} catch (...) {
	  unlock ();
	  throw;
	}
      }
      if (quark == QUARK_ADDHEAD) {
	Object* obj = argv->get (0);
	HtmlTh* th = dynamic_cast <HtmlTh*> (obj);
	if (th != nilp) {
	  HtmlTableData* result = adddata (th);
	  robj->post (result);
	  unlock ();
	  return result;
	}
	Literal* lval = dynamic_cast <Literal*> (obj);
	if (lval != nilp) {
	  HtmlTableData* result = addhead (*lval);
	  robj->post (result);
	  unlock ();
	  return result;
	}
	throw Exception ("type-error", "unexpected object with add-head",
			 Object::repr (obj));
      }
      if (quark == QUARK_GETTD) {
	rdlock ();
	try {
	  long index = argv->getint (0);
	  HtmlTableData* td = gettd (index);
	  robj->post (td);
	  return td;
	} catch (...) {
	  unlock ();
	  throw;
	}
      }
      if (quark == QUARK_SETDTDC) {
	String dtdc = argv->getstring (0);
	setdtdc (dtdc);
	return nilp;
      }
    }
    // call the html element method
    return HtmlElement::apply (robj, nset, quark, argv);
  }
}
